"""
Merge paired-end reads (a wrapper around PEAR)

This script can also manage a cache of already-merged files.
"""
import os
import sys
from pathlib import Path
import logging
import tempfile
import subprocess
import hashlib
import shutil

from igdiscover.config import GlobalConfig

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--threads', '-j', default=None, type=int, help='Number of threads')
	arg('--no-cache', default=None, action='store_true',
		help='Disable cache. Default: Determined by configuration')
	arg('reads1', help='Forward reads FASTQ file')
	arg('reads2', help='Reverse reads FASTQ file')
	arg('output',  help='Output file (compressed FASTQ)')


def compute_hash(path1: str, path2: str):
	result = subprocess.run(['pear', '-h'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	pear_version = None
	for line in result.stdout.split(b'\n'):
		if line.startswith(b'PEAR'):
			pear_version = line
			break
	assert pear_version is not None

	hasher = hashlib.md5(pear_version)
	for path in path1, path2:
		with open(path, 'rb') as f:
			while True:
				chunk = f.read(1048576)
				if not chunk:
					break
				hasher.update(chunk)
	return hasher.hexdigest()


def run_pear(path1: str, path2: str, output: str, log_output=None, threads: int=None):
	gzip = 'pigz' if shutil.which('pigz') else 'gzip'
	with tempfile.TemporaryDirectory() as tmpdir:
		tmpdir = Path(tmpdir)
		args = ['pear']
		if threads is not None:
			args += ['-j', str(threads)]
		args += [
			'-f', path1,
			'-r', path2,
			'-o', tmpdir / 'merged'
		]
		if log_output is None:
			subprocess.run(args, check=True)
		else:
			result = subprocess.run(args, check=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
			with open(log_output, 'wb') as f:
				f.write(result.stdout)

		subprocess.run([gzip, tmpdir / 'merged.assembled.fastq'], check=True)
		shutil.move(tmpdir / 'merged.assembled.fastq.gz', output)


def run_pear_cached(path1: str, path2: str, output: str, threads: int = None):
	cache_home = Path(os.environ.get('XDG_CACHE_HOME', os.path.expanduser('~/.cache')))
	cachedir = cache_home / 'igdiscover' / 'mergedreads'
	cachedir.mkdir(parents=True, exist_ok=True)

	md5sum = compute_hash(path1, path2)
	cached_merged = cachedir / (md5sum + '.fastq.gz')
	cached_log = cachedir / (md5sum + '.log')
	if not cached_merged.exists() or not cached_log.exists():
		run_pear(path1, path2, output=cached_merged, log_output=cached_log, threads=threads)
		logger.info('PEAR result copied to cache')
	else:
		logger.info('PEAR result found in cache:')
		logger.info(cached_merged)
		logger.info(cached_log)
	shutil.copy(cached_merged, output)
	with open(cached_log) as f:
		sys.stdout.write(f.read())

	# Update mtime of the cache file. Let’s us find unused cache entries.
	for path in cached_merged, cached_log:
		try:
			path.touch()
		except PermissionError:
			pass


def main(args):
	use_cache = GlobalConfig().use_cache
	if args.no_cache:
		use_cache = False
	if use_cache:
		logger.info('Cache enabled')
		func = run_pear_cached
	else:
		func = run_pear
	func(path1=args.reads1, path2=args.reads2, output=args.output, threads=args.threads)



#!/usr/bin/env python3
"""
Bubble plot
"""
import logging
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib
import seaborn as sns
import pandas as pd
import numpy as np

#sns.set(style='white', font_scale=1.5, rc={"lines.linewidth": 1})
logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--scale', default=1.0, type=float, help='scaling factor for bubble size (default %(default)s)')
	arg('pdf', help='PDF output')
	arg('table', help='Input table')


def main(args):
	with open(args.table) as f:
		# Try to auto-detect the separator
		if ';' in f.read():
			kwargs = {'sep': ';'}
		else:
			kwargs = {}
	df = pd.read_table(args.table, index_col=0, **kwargs)
	m = len(df.index)
	n_compartments = len(df.columns)
	logger.info('Table with %s rows read', len(df))
	logger.info('%s compartments', n_compartments)
	df = df.unstack().reset_index()
	df.columns = ['compartment', 'clone', 'size']
	colors = {
		"BLOOD": "#990000",
		"BM": "#0000CC",
		"SPLEEN": "#336600",
		"LN": "#999999",
		"GUT": "#660099"
	}

	fig = Figure(figsize=((n_compartments+2)*.6, (m+1)*.4))#, sharex=True, sharey=True)
	FigureCanvas(fig)
	ax = fig.add_subplot(111)

	sns.scatterplot(
		data=df, y='clone', x='compartment', hue='compartment', size='size',
		sizes=(0, args.scale * 1000), palette=colors.values(), ax=ax)

	ax.set_xlabel('Traced lineages')
	ax.set_ylabel('')
	ax.set_ylim(-1, m)
	#ax.xaxis.set_tick_params(rotation=90)
	ax.grid(axis='y', linestyle=':')
	ax.set_axisbelow(True)

	handles, labels = ax.get_legend_handles_labels()
	handles = handles[2+n_compartments:]
	labels = labels[2+n_compartments:]
	ax.legend(
		handles, labels, bbox_to_anchor=(1.1, 0.55),
		loc=6, labelspacing=3, borderaxespad=0.,
		handletextpad=1,
		frameon=False)#, title='')

	fig.savefig(args.pdf, bbox_inches='tight')
	logger.info('File %r written', args.pdf)


if __name__ == '__main__':
	logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
	from argparse import ArgumentParser
	parser = ArgumentParser()
	add_arguments(parser)
	args = parser.parse_args()
	main(args)



"""
Query two ...
"""
import sys
import logging
from itertools import islice
from collections import Counter
from io import StringIO
from matplotlib.backends.backend_pdf import FigureCanvasPdf, PdfPages
import pandas as pd
import seaborn as sns

sns.set(style='white', font_scale=1.5, rc={"lines.linewidth": 1})
logger = logging.getLogger(__name__)
CM = 1 / 2.54


def add_arguments(parser):
	arg = parser.add_argument
	arg('--nt', default=False, action='store_true',
		help='Count CDR3 mismatches on nucleotide level. Default: Compare amino-acids.')
	arg('--limit', default=None, type=int, metavar='N',
		help='Read only the first N clonotypes of each table (for fast testing)')
	arg('--minsize', default=5, type=int, metavar='N',
		help='Require at least N members in each dataset to plot a clonotype')
	arg('--names', metavar='NAME[,NAME]...',
		help='Comma-separated list of dataset names to be used in the legend')
	arg('pdf', help='PDF output')
	arg('tables', nargs='+', help='clonotype member tables')


def read_dataset(path, cdr3_column='CDR3_aa', limit=None, minsize=1):
	usecols = [
		'V_gene',
		'J_gene',
		cdr3_column,
		'FR1_SHM',
		'CDR1_SHM',
		'FR2_SHM',
		'CDR2_SHM',
		'FR3_SHM',
		'V_SHM',
		'J_SHM',
	]
	column_names = list(pd.read_csv(path, sep='\t', nrows=0).columns)
	assert 'CDR3_length' in column_names
	logger.info('Reading %s', path)
	with open(path) as f:
		data = f.read()
	first = True
	tables = []
	for i, chunk in enumerate(islice(data.split('\n\n'), limit), 1):
		if not chunk:
			# The file ends with an empty line that needs to be skipped
			continue
		table = pd.read_csv(StringIO(chunk), sep='\t', usecols=usecols, header=0 if first else None,
			names=column_names)
		first = False
		assert len(set(table[cdr3_column])) == 1
		if len(table) >= minsize:
			tables.append(table.set_index(['V_gene', 'J_gene', cdr3_column]).sort_index())
		if i % 10000 == 0:
			logger.info('Read %s clones', i)
	logger.info('Read %s clones', i)
	return pd.concat(tables)


def main(args):
	logger.info('Will plot results to %s', args.pdf)
	cdr3_column = 'CDR3_nt' if args.nt else 'CDR3_aa'
	n_datasets = len(args.tables)
	if args.names:
		names = args.names.split(',')
		if len(names) != n_datasets:
			logger.error('%s dataset names given, but %s datasets provided', len(names), n_datasets)
			sys.exit(1)
	else:
		names = list(range(n_datasets))

	datasets = (read_dataset(path, limit=args.limit, minsize=args.minsize) for path in args.tables)
	df = pd.concat(datasets, keys=range(n_datasets), names=['dataset_id'])
	logger.info('Read %s tables', n_datasets)
	df.rename(columns=lambda x: x[:-4], inplace=True)  # Removes _SHM suffix
	cols = ['V_gene', 'J_gene', cdr3_column]
	n = 0
	with PdfPages(args.pdf) as pages:
		for (v_gene, j_gene, cdr3), group in df.groupby(level=cols):
			group = group.reset_index(level=cols, drop=True)
			skip = False
			counter = Counter(group.index)
			for dataset_id in range(n_datasets):
				if counter[dataset_id] < args.minsize:
					skip = True
					break
			if skip:
				continue
			table = group.stack()
			table.index.set_names('region', level=1, inplace=True)
			table.name = 'SHM'
			table = table.reset_index()
			table = table.assign(Dataset=table['dataset_id'].map(lambda i: names[i]))
			g = sns.factorplot(data=table, x='region', y='SHM', hue='Dataset',
				kind='violin', size=16*CM, aspect=2)
			dscounts = ' vs '.join(str(counter[i]) for i in range(n_datasets))
			g.fig.suptitle('V: {} – J: {} – CDR3: {} ({})'.format(v_gene, j_gene, cdr3, dscounts))
			g.set_axis_labels('Region')
			g.set_ylabels('%SHM')
			# g.despine()
			FigureCanvasPdf(g.fig).print_figure(pages, bbox_inches='tight')
			n += 1
			logger.info('Plotted %s clonotypes', n)


if __name__ == '__main__':
	logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
	from argparse import ArgumentParser
	parser = ArgumentParser()
	add_arguments(parser)
	args = parser.parse_args()
	main(args)



"""
Determine haplotypes based on co-occurrences of alleles
"""
import sys
import logging
from typing import List, Tuple, Iterator
from itertools import product
import pandas as pd
from argparse import ArgumentParser
from sqt import SequenceReader
from .table import read_table

logger = logging.getLogger(__name__)


# The second-most expressed allele of a gene must be expressed at at least this
# fraction of the highest-expressed allele in order for the gene to be considered
# heterozygous.
HETEROZYGOUS_THRESHOLD = 0.1

#
EXPRESSED_RATIO = 0.1


def add_arguments(parser: ArgumentParser):
	arg = parser.add_argument
	arg('--v-gene', help='V gene to use for haplotyping J. Default: Auto-detected')
	arg('--d-evalue', type=float, default=1E-4,
		help='Maximal allowed E-value for D gene match. Default: %(default)s')
	arg('--d-coverage', '--D-coverage', type=float, default=65,
		help='Minimum D coverage (in percent). Default: %(default)s%%)')
	arg('--restrict', metavar='FASTA',
		help='Restrict analysis to the genes named in the FASTA file. '
			'Only the sequence names are used!')
	arg('--order', metavar='FASTA', default=None,
		help='Sort the output according to the order of the records in '
			'the given FASTA file.')
	arg('--plot', metavar='FILE', default=None,
		help='Write a haplotype plot to FILE')
	arg('--structure-plot', metavar='FILE', default=None,
		help='Write a haplotype structure plot (counts binarized 0 and 1) to FILE')
	arg('table', help='Table with parsed and filtered IgBLAST results')


def expression_counts(table: pd.DataFrame, gene_type: str) -> Iterator[pd.DataFrame]:
	"""
	Yield DataFrames for each gene with gene and allele as the row index and columns 'name' and
	'count'. For example, when 'name' is VH1-1*01, gene would be 'VH1-1' and allele
	would be '01'.
	"""
	counts = table.groupby(gene_type + '_gene').size()
	names, _, alleles = zip(*[s.partition('*') for s in counts.index])
	expressions = pd.DataFrame(
		{'gene': names, 'allele': alleles, 'count': counts, 'name': counts.index},
		columns=['gene', 'allele', 'name', 'count']).set_index(['gene', 'allele'])
	del alleles

	# Example expressions at this point:
	#
	#                             name  count
	# gene       allele
	# IGHV1-18   01        IGHV1-18*01    166
	#            03        IGHV1-18*03      1
	# IGHV1-2    02         IGHV1-2*02     85
	#            04         IGHV1-2*04     16
	# IGHV1-24   01        IGHV1-24*01      5

	logger.info('Heterozygous %s genes:', gene_type)
	for _, alleles in expressions.groupby(level='gene'):
		# Remove alleles that have too low expression relative to the highest-expressed allele
		max_expression = alleles['count'].max()
		alleles = alleles[alleles['count'] >= HETEROZYGOUS_THRESHOLD * max_expression]
		if len(alleles) >= 2:
			logger.info('%s with alleles %s -- Counts: %s',
				alleles.index[0][0],
				', '.join(alleles['name']),
				', '.join(str(x) for x in alleles['count'])
			)
		yield alleles


class HeterozygousGene:
	def __init__(self, name: str, alleles: List[str]):
		"""
		name -- name of this gene, such as 'VH4-4'
		alleles -- list of its alleles, such as ['VH4-4*01', 'VH4-4*02']
		"""
		self.name = name
		self.alleles = alleles


def compute_coexpressions(table: pd.DataFrame, gene_type1: str, gene_type2: str):
	assert gene_type1 != gene_type2
	coexpressions = table.groupby(
		(gene_type1 + '_gene', gene_type2 + '_gene')).size().to_frame()
	coexpressions.columns = ['count']
	return coexpressions


def cooccurrences(coexpressions, het_alleles: Tuple[str, str], target_groups):
	"""
	het_alleles -- a pair of alleles of a heterozygous gene,
	such as ('IGHJ6*02', 'IGHJ6*03').
	"""
	assert len(het_alleles) == 2

	haplotype = []
	for target_alleles in target_groups:
		is_expressed_list = []
		names = []
		counts = []
		for target_allele, _ in target_alleles.itertuples(index=False):
			ex = []
			for het_allele in het_alleles:
				try:
					e = coexpressions.loc[(het_allele, target_allele), 'count']
				except KeyError:
					e = 0
				ex.append(e)
			ex_total = sum(ex) + 1  # +1 avoids division by zero
			ratios = [x / ex_total for x in ex]
			is_expressed = [ratio >= EXPRESSED_RATIO for ratio in ratios]
			if is_expressed != [False, False]:
				is_expressed_list.append(is_expressed)
				names.append(target_allele)
				counts.append(ex)
		if len(is_expressed_list) == 1:
			is_expressed = is_expressed_list[0]
			if is_expressed == [True, False]:
				haplotype.append((names[0], '', 'deletion', counts[0]))
			elif is_expressed == [False, True]:
				haplotype.append(('', names[0], 'deletion', counts[0]))
			elif is_expressed == [True, True]:
				haplotype.append((names[0], names[0], 'homozygous', counts[0]))
			else:
				assert False
		elif is_expressed_list == [[True, False], [False, True]]:
			haplotype.append((names[0], names[1], 'heterozygous', (counts[0][0], counts[1][1])))
		elif is_expressed_list == [[False, True], [True, False]]:
			haplotype.append((names[1], names[0], 'heterozygous', (counts[0][1], counts[1][0])))
		else:
			type_ = 'unknown'
			# Somewhat arbitrary criteria for a "duplication":
			# 1) one heterozygous allele, 2) at least three alleles in total
			n_true = sum(x.count(True) for x in is_expressed_list)
			if ([True, False] in is_expressed_list or [False, True] in is_expressed_list) and n_true > 2:
				type_ = 'duplication'
			for is_expressed, name, count in zip(is_expressed_list, names, counts):
				haplotype.append((
					name if is_expressed[0] else '',
					name if is_expressed[1] else '',
					type_,
					count,
				))
	return haplotype


class HaplotypePair:
	"""Haplotype pair for a single gene type (V/D/J)"""

	def __init__(self, haplotype, gene_type, het1, het2):
		self.haplotype = haplotype
		self.gene_type = gene_type
		self.het1 = het1
		self.het2 = het2

	def sort(self, order: List[str]) -> None:
		"""
		Sort the haplotype
		order -- list a names of genes in the desired order
		"""
		gene_order = {name: i for i, name in enumerate(order)}

		def keyfunc(hap):
			name = hap[0] if hap[0] else hap[1]
			gene, _, allele = name.partition('*')
			try:
				allele = int(allele)
			except ValueError:
				allele = 999
			try:
				index = gene_order[gene]
			except KeyError:
				logger.warning('Gene %s not found in gene order file, placing it at the end',
					gene)
				index = 1000000
			return index * 1000 + allele

		self.haplotype = sorted(self.haplotype, key=keyfunc)

	def switch(self):
		"""Swap the two haplotypes"""
		self.het2, self.het1 = self.het1, self.het2
		haplotype = []
		for name1, name2, type_, counts in self.haplotype:
			assert len(counts) == 2
			counts = counts[1], counts[0]
			haplotype.append((name2, name1, type_, counts))
		self.haplotype = haplotype

	def to_tsv(self, header: bool=True) -> str:
		lines = []
		if header:
			lines.append('\t'.join(['haplotype1', 'haplotype2', 'type', 'count1', 'count2']))
		lines.append(
			'# {} haplotype from {} and {}'.format(self.gene_type, self.het1, self.het2))
		for h1, h2, typ, count in self.haplotype:
			lines.append('\t'.join([h1, h2, typ, str(count[0]), str(count[1])]))
		return '\n'.join(lines) + '\n'


def plot_haplotypes(blocks: List[HaplotypePair], show_unknown: bool=False, binarize: bool=False):
	from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
	from matplotlib.figure import Figure
	from matplotlib.patches import Patch

	colormap = dict(homozygous='cornflowerblue', heterozygous='lightgreen', deletion='gold',
		duplication='crimson', unknown='gray')
	if not show_unknown:
		del colormap['unknown']
	labels = [[], []]
	heights = [[], []]
	names = []
	colors = []
	positions = []
	pos = 0
	for block in blocks:
		for hap1, hap2, type_, (count1, count2) in block.haplotype:
			if not show_unknown and type_ == 'unknown':
				continue
			if binarize:
				count1 = 1 if hap1 else 0
				count2 = 1 if hap2 else 0
			label = hap1 if hap1 else hap2
			hap1 = ('*' + hap1.partition('*')[2]) if hap1 else ''
			hap2 = ('*' + hap2.partition('*')[2]) if hap2 else ''
			heights[0].append(count1)
			heights[1].append(count2)
			labels[0].append(hap1)
			labels[1].append(hap2)
			names.append(label.partition('*')[0])
			colors.append(colormap[type_])
			positions.append(pos)
			pos += 1
		pos += 2

	n = len(names)
	assert len(labels[0]) == len(labels[1]) == len(colors) == len(heights[0]) == len(heights[1]) == n

	fig = Figure(figsize=(8, 24))
	FigureCanvas(fig)
	axes = fig.subplots(ncols=2)
	for i in 0, 1:
		axes[i].barh(y=positions, width=heights[i], color=colors)
		axes[i].set_yticklabels(labels[i])
		axes[i].set_ylim((-0.5, max(positions) + 0.5))

	# Add center axis
	ax_center = axes[0].twinx()
	ax_center.set_yticks(axes[0].get_yticks())
	ax_center.set_ylim(axes[0].get_ylim())
	ax_center.set_yticklabels(names)

	# Synchronize x limits on both axes (has no effect if binarize is True)
	max_x = max(axes[i].get_xlim()[1] for i in range(2))
	for ax in axes:
		ax.set_xlim(right=max_x)
	axes[0].invert_xaxis()

	for ax in axes[0], axes[1], ax_center:
		ax.invert_yaxis()
		for spine in ax.spines.values():
			spine.set_visible(False)
		if binarize:
			ax.set_xticks([])
		else:
			ax.set_axisbelow(True)
			ax.grid(True, axis='x')
		ax.set_yticks(positions)
		ax.tick_params(left=False, right=False, labelsize=12)
	axes[1].tick_params(labelleft=False, labelright=True)
	if not binarize:
		axes[0].spines['right'].set_visible(True)
		axes[1].spines['left'].set_visible(True)

	# Legend
	legend_patches = [Patch(color=col, label=label) for label, col in colormap.items()]
	fig.legend(handles=legend_patches, loc='upper center', ncol=len(legend_patches))
	fig.tight_layout()
	fig.subplots_adjust(top=1 - 2 / len(names))

	return fig


def read_and_filter(path: str, d_evalue: float, d_coverage: float):
	usecols = ['V_gene', 'D_gene', 'J_gene', 'V_errors', 'D_errors', 'J_errors', 'D_covered',
		'D_evalue']
	# Support reading a table without D_errors
	try:
		table = read_table(path, usecols=usecols)
	except ValueError:
		usecols.remove('D_errors')
		table = read_table(path, usecols=usecols)
	logger.info('Table with %s rows read', len(table))

	table = table[table.V_errors == 0]
	logger.info('%s rows remain after requiring V errors = 0', len(table))
	table = table[table.J_errors == 0]
	logger.info('%s rows remain after requiring J errors = 0', len(table))
	table = table[table.D_evalue <= d_evalue]
	logger.info('%s rows remain after requiring D E-value <= %s', len(table), d_evalue)
	table = table[table.D_covered >= d_coverage]
	logger.info('%s rows remain after requiring D coverage >= %s', len(table), d_coverage)
	if 'D_errors' in table.columns:
		table = table[table.D_errors == 0]
		logger.info('%s rows remain after requiring D errors = 0', len(table))
	return table


def main(args):
	if args.order is not None:
		with SequenceReader(args.order) as sr:
			gene_order = [r.name for r in sr]
	else:
		gene_order = None

	table = read_and_filter(args.table, args.d_evalue, args.d_coverage)

	if args.restrict is not None:
		with SequenceReader(args.restrict) as sr:
			restrict_names = set(r.name for r in sr)
		table = table[table['V_gene'].map(lambda name: name in restrict_names)]
		logger.info('After restricting to V genes named in %r, %d rows remain', args.restrict,
			len(table))
		if len(table) == 0:
			logger.error('No rows remain, cannot continue')
			sys.exit(1)

	expressions = dict()
	het_expressions = dict()  # these are also sorted, most highly expressed first
	for gene_type in 'VDJ':
		ex = list(expression_counts(table, gene_type))
		expressions[gene_type] = ex
		het_ex = [e for e in ex if len(e) == 2]
		if het_ex:
			# Pick most highly expressed
			het_expressions[gene_type] = sorted(het_ex, key=lambda e: e['count'].sum(), reverse=True)[:5]
		else:
			# Force at least something to be plotted
			het_expressions[gene_type] = [None]

	if args.v_gene:
		het_ex = [e for e in expressions['V'] if len(e) == 2]
		for ex in het_ex:
			if (args.v_gene in ex.index and not ex.loc[args.v_gene].empty) or (args.v_gene in ex['name'].values):
				het_expressions['V'] = [ex]
				break
		else:
			logger.error('The gene or allele %s was not found in the list of heterozygous V genes. '
				'It cannot be used with the --v-gene option.', args.v_gene)
			sys.exit(1)

	block_lists = []

	# We want to avoid using a gene classifed as 'duplicate' for haplotyping, but the
	# classification is only known after we have done the haplotyping, so we try it
	# until we found a combination that works
	products = list(product(het_expressions['J'], het_expressions['V']))
	for attempt, (het_j, het_v) in enumerate(products):
		best_het_genes = {
			'V': het_v,
			'D': het_expressions['D'][0] if het_expressions['D'] else None,
			'J': het_j,
		}
		for gene_type in 'VDJ':
			bhg = best_het_genes[gene_type]
			text = bhg.index[0][0] if bhg is not None else 'none found'
			logger.info('Heterozygous %s gene to use for haplotyping: %s',
				gene_type, text)

		# Create HaplotypePair objects ('blocks') for each gene type
		blocks = []

		for target_gene_type, het_gene in (
			('J', 'V'),
			('D', 'J'),
			('V', 'J'),
		):
			het_alleles = best_het_genes[het_gene]
			if het_alleles is None:
				continue
			coexpressions = compute_coexpressions(table, het_gene, target_gene_type)
			target_groups = expressions[target_gene_type]
			het1, het2 = het_alleles['name']
			haplotype = cooccurrences(coexpressions, (het1, het2), target_groups)
			block = HaplotypePair(haplotype, target_gene_type, het1, het2)
			if gene_order:
				block.sort(gene_order)
			blocks.append(block)

		if het_j is None or het_v is None:
			break
		het_used = set(sum([list(h['name']) for h in best_het_genes.values() if h is not None], []))
		het_is_duplicate = False
		for block in blocks:
			if block.gene_type == 'D':
				continue

			# This nested loop is put in a separate generator so we can easily 'break'
			# out of both loops at the same time.
			def nameiter():
				for name1, name2, type_, _ in block.haplotype:
					for name in name1, name2:
						yield name, type_

			for name, type_ in nameiter():
				if name in het_used and type_ != 'heterozygous':
					het_is_duplicate = True
					if not args.v_gene:
						logger.warning('%s not classified as "heterozygous" during haplotyping, '
							'attempting to use different alleles', name)
					break
			if het_is_duplicate:
				break
		block_lists.append(blocks)
		if not het_is_duplicate:
			break
	else:
		if not args.v_gene:
			logger.warning('No other alleles remain, using first found solution')
		blocks = block_lists[0]

	# Get the phasing right across blocks (i.e., swap J haplotypes if necessary)
	assert len(blocks) in (0, 1, 3)
	if len(blocks) == 3:
		j_hap, d_hap, v_hap = blocks
		assert j_hap.gene_type == 'J'
		assert d_hap.gene_type == 'D'
		assert v_hap.gene_type == 'V'
		assert d_hap.het1 == v_hap.het1
		assert d_hap.het2 == v_hap.het2
		for name1, name2, _, _ in j_hap.haplotype:
			if (name1, name2) == (v_hap.het2, v_hap.het1):
				j_hap.switch()
				break

	# Print the table
	header = True
	for block in blocks:
		print(block.to_tsv(header=header))
		header = False

	# Create plots if requested
	if args.plot:
		fig = plot_haplotypes(blocks, show_unknown=True)
		fig.savefig(args.plot)
	if args.structure_plot:
		fig = plot_haplotypes(blocks, binarize=True)
		fig.savefig(args.structure_plot)



#!/usr/bin/env python3
"""
IgDiscover computes V/D/J gene usage profiles and discovers novel V genes

- Run IgBLAST in parallel (wrapper inspired by igblastwrp).
- Parse IgBLAST output into a tab-separated table
- Group sequences by barcode
- Plot V gene usage
- Discover new V genes given more than one dataset
"""
import sys
import logging
import importlib
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import matplotlib as mpl
import warnings
import resource
from . import __version__

__author__ = "Marcel Martin"

mpl.use('Agg')
warnings.filterwarnings('ignore', 'axes.color_cycle is deprecated and replaced with axes.prop_cycle')
warnings.filterwarnings('ignore', 'The `IPython.html` package')

# List of all subcommands. A module of the given name must exist and define
# add_arguments() and main() functions. Documentation is taken from the first
# line of the module’s docstring.
COMMANDS = [
	'init',
	'run',
	'config',
	'commonv',
	'igblast',
	'filter',
	'count',
	'group',
	'dereplicate',
	#'multidiscover',
	'germlinefilter',
	'discover',
	'discoverj',
	'clusterplot',
	'errorplot',
	'upstream',
	'dendrogram',
	'rename',
	'union',
	'dbdiff',
	'merge',
	'clonotypes',
	'clonoquery',
	'plotalleles',
	'haplotype',
]

logger = logging.getLogger(__name__)


class HelpfulArgumentParser(ArgumentParser):
	"""An ArgumentParser that prints full help on errors."""

	def __init__(self, *args, **kwargs):
		if 'formatter_class' not in kwargs:
			kwargs['formatter_class'] = RawDescriptionHelpFormatter
		super().__init__(*args, **kwargs)

	def error(self, message):
		self.print_help(sys.stderr)
		args = {'prog': self.prog, 'message': message}
		self.exit(2, '%(prog)s: error: %(message)s\n' % args)


def format_duration(seconds):
	h = int(seconds // 3600)
	seconds -= h * 3600
	m = int(seconds // 60)
	seconds -= m * 60
	return '{:02d}:{:02d}:{:04.1f}'.format(h, m, seconds)


def main(arguments=None):
	logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
	parser = HelpfulArgumentParser(description=__doc__, prog='igdiscover')
	parser.add_argument('--profile', default=False, action='store_true',
		help='Save profiling information to igdiscover.prof')
	parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

	show_cpustats = dict()
	subparsers = parser.add_subparsers()
	for command_name in COMMANDS:
		module = importlib.import_module('.' + command_name, 'igdiscover')
		subparser = subparsers.add_parser(command_name,
			help=module.__doc__.split('\n')[1], description=module.__doc__)
		subparser.set_defaults(func=module.main)
		module.add_arguments(subparser)
		if hasattr(module, 'do_not_show_cpustats'):
			show_cpustats[module.main] = False

	args = parser.parse_args(arguments)
	if not hasattr(args, 'func'):
		parser.error('Please provide the name of a subcommand to run')
	elif args.profile:
		import cProfile as profile
		profile.runctx('args.func(args)', globals(), locals(), filename='igdiscover.prof')
		logger.info('Wrote profiling data to igdiscover.prof')
	else:
		args.func(args)
	if sys.platform == 'linux' and show_cpustats.get(args.func, True):
		rself = resource.getrusage(resource.RUSAGE_SELF)
		rchildren = resource.getrusage(resource.RUSAGE_CHILDREN)
		memory_kb = rself.ru_maxrss + rchildren.ru_maxrss
		cpu_time = rself.ru_utime + rself.ru_stime + rchildren.ru_utime + rchildren.ru_stime
		cpu_time_s = format_duration(cpu_time)
		logger.info('CPU time {}. Maximum memory usage {:.3f} GB'.format(
			cpu_time_s, memory_kb / 1E6))


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Compute and plot V/J gene usage from a clonotypes.tab file
"""
import sys
from argparse import ArgumentParser
import seaborn
import pandas as pd


def add_arguments(parser):
	arg = parser.add_argument
	arg('--mincount', metavar='N', default=1, type=int,
		help='Filter out clonotypes with less than N members. Default: %(default)s')
	arg('--gene', default='V', choices=('V', 'J'),
		help='Which gene to plot. Choose V or J. Default: %(default)s')
	arg('--plot', metavar='FILE',
		help='Plot results into FILE (PNG, PDF)')
	arg('table', help='clonotypes.tab table created with the "clonotypes" subcommand')


def main(args):
	# Read in input table
	table = pd.read_csv(args.table, sep='\t')

	# Extract the 'V_gene' or 'J_gene' column
	counts_column = table[args.gene + '_gene']

	# Count how often each value (gene) appears
	gene_frequencies = counts_column.value_counts()

	# Keep only those with minimum count
	gene_frequencies = gene_frequencies[gene_frequencies >= args.mincount]

	# Sort by gene name
	gene_frequencies = gene_frequencies.sort_index()

	# Print out the frequencies
	print('gene', 'frequency', sep='\t')
	for gene_name, frequency in gene_frequencies.items():
		print(gene_name, frequency, sep='\t')

	# If requested, create a plot
	if args.plot:
		ax = gene_frequencies.plot(kind='bar', figsize=(20, 5))
		ax.set_title('Gene expression counts')
		ax.set_xlabel('{} gene'.format(args.gene))
		ax.set_ylabel('Clonotype count')
		ax.figure.tight_layout()
		ax.figure.savefig(args.plot)



if __name__ == '__main__':
	parser = ArgumentParser(description=__doc__)
	add_arguments(parser)
	args = parser.parse_args()
	main(args)



#!/usr/bin/env python3
import sys
import subprocess


def download_sra(sra_id):
	if not sra_id.startswith('SRR') and not sra_id.startswith('ERR'):
		raise ValueError('only ids starting with SRR or ERR supported')
	number = sra_id[3:]

	if len(number) == 6:
		url = f'ftp://ftp.sra.ebi.ac.uk/vol1/fastq/{sra_id[:6]}/{sra_id}/'
	elif len(number) == 7:
		url = f'ftp://ftp.sra.ebi.ac.uk/vol1/fastq/{sra_id[:6]}/00{number[6]}/{sra_id}/'
	else:
		raise ValueError('number not supported')

	single_end_url = f'{url}{sra_id}.fastq.gz'
	paired_end_urls = [f'{url}{sra_id}_{i}.fastq.gz' for i in (1, 2)]

	result = subprocess.run(['wget', single_end_url])
	if result.returncode != 0:
		for url in paired_end_urls:
			subprocess.run(['wget', url], check=True)


if __name__ == '__main__':
	download_sra(sys.argv[1])



"""
Compare two FASTA files based on sequences

The order of records in the two files does not matter.

Exit code:
    2 if duplicate sequences or duplicate record names were found
    1 if there are any lost or gained records or sequence differences
    0 if the records are identical, but allowing for different record names

"""
import sys
import logging
import numpy as np
from scipy.optimize import linear_sum_assignment
from sqt import FastaReader
from sqt.align import hamming_distance

logger = logging.getLogger(__name__)

do_not_show_cpustats = 1


def add_arguments(parser):
	arg = parser.add_argument
	arg('--color', default='auto', choices=('auto', 'never', 'always'),
		help='Whether to colorize output')
	arg('a', help='FASTA file with expected sequences')
	arg('b', help='FASTA file with actual sequences')


RED = "\x1b[0;31m"
GREEN = "\x1b[0;32m"
RESET = "\x1b[0m"


def red(s):
	return RED + s + RESET


def green(s):
	return GREEN + s + RESET


def check_duplicate_names(records):
	names = set()
	for record in records:
		if record.name in names:
			yield record.name
		names.add(record.name)


def check_exact_duplicate_sequences(records):
	sequences = dict()
	for record in records:
		if record.sequence in sequences:
			yield record.name, sequences[record.sequence]
		else:
			sequences[record.sequence] = record.name


def compare(a, b):
	"""Return cost of comparing a to b"""

	l = min(len(a.sequence), len(b.sequence))
	length_diff = max(len(a.sequence), len(b.sequence)) - l
	dist_prefixes = hamming_distance(a.sequence[:l], b.sequence[:l])
	dist_suffixes = hamming_distance(a.sequence[-l:], b.sequence[-l:])

	return 5 * min(dist_prefixes, dist_suffixes) + length_diff


def pair_up_identical(a_records, b_records):
	identical = []
	b_map = {record.sequence: record for record in b_records}
	a_rest = []
	for a in a_records:
		if a.sequence in b_map:
			identical.append((a, b_map[a.sequence]))
			del b_map[a.sequence]
		else:
			a_rest.append(a)

	return identical, a_rest, list(b_map.values())


def pair_up(a_records, b_records, max_cost=20):
	# Pair up identical sequences first
	identical, a_records, b_records = pair_up_identical(a_records[:], b_records[:])

	# Compare all vs all and fill in a score matrix
	m = len(a_records)
	n = len(b_records)
	cost = np.zeros((m, n), dtype=int)
	for i, a in enumerate(a_records):
		for j, b in enumerate(b_records):
			cost[i, j] = compare(a, b)

	# Solve minimum weighted bipartite matching
	assignment = linear_sum_assignment(cost)
	similar = []
	a_similar = set()
	b_similar = set()
	for i, j in zip(*assignment):
		if cost[i, j] <= max_cost:
			similar.append((a_records[i], b_records[j]))
			a_similar.add(i)
			b_similar.add(j)

	a_only = [a for i, a in enumerate(a_records) if i not in a_similar]
	b_only = [b for j, b in enumerate(b_records) if j not in b_similar]

	return a_only, b_only, identical, similar


def format_indel(a, b, colored: bool):
	if len(a) > len(b):
		assert len(b) == 0
		s = '{-' + a + '}'
		return red(s) if colored else s
	elif len(b) > len(a):
		assert len(a) == 0
		s = '{+' + b + '}'
		return green(s) if colored else s
	else:
		return ''


def print_similar(a, b, colored: bool):
	l = min(len(a.sequence), len(b.sequence))
	dist_prefixes = hamming_distance(a.sequence[:l], b.sequence[:l])
	dist_suffixes = hamming_distance(a.sequence[-l:], b.sequence[-l:])
	if dist_prefixes <= dist_suffixes:
		a_prefix = ''
		b_prefix = ''
		a_common = a.sequence[:l]
		b_common = b.sequence[:l]
		a_suffix = a.sequence[l:]
		b_suffix = b.sequence[l:]
	else:
		a_prefix = a.sequence[:-l]
		b_prefix = b.sequence[:-l]
		a_common = a.sequence[-l:]
		b_common = b.sequence[-l:]
		a_suffix = ''
		b_suffix = ''

	s = format_indel(a_prefix, b_prefix, colored)
	edits = []
	for i, (ac, bc) in enumerate(zip(a_common, b_common)):
		if ac != bc:
			if colored:
				s = '{' + red(ac) + ' → ' + green(bc) + '}'
			else:
				s = '{' + ac + ' → ' + bc + '}'
			edits.append(s)
		else:
			edits.append(ac)
	s += ''.join(edits)

	s += format_indel(a_suffix, b_suffix, colored)

	print('~', a.name, '--', b.name)
	print(s)
	print()


def main(args):
	if args.color == 'auto':
		colored = sys.stdout.isatty()
	elif args.color == 'never':
		colored = False
	else:
		assert args.color == 'always'
		colored = True

	with FastaReader(args.a) as f:
		a_records = list(f)
	with FastaReader(args.b) as f:
		b_records = list(f)

	has_duplicate_names = False
	for records, path in ((a_records, args.a), (b_records, args.b)):
		dups = list(check_duplicate_names(records))
		if dups:
			has_duplicate_names = True
			print('Duplicate record names found in', path)
			for name in dups:
				print('-', name)

	has_duplicate_sequences = False
	for record, path in ((a_records, args.a), (b_records, args.b)):
		dups = list(check_exact_duplicate_sequences(records))
		if dups:
			has_duplicate_sequences = True
			print('Duplicate sequences found in', path)
			for name, name_orig in dups:
				print('-', name, 'is identical to earlier record', name_orig)

	only_a, only_b, identical, similar = pair_up(a_records, b_records)
	different_name = [(a, b) for a, b in identical if a.name != b.name]

	# Summary
	print('{} vs {} records. {} lost, {} gained, {} identical, {} different name, {} similar'.format(
		len(a_records), len(b_records), len(only_a), len(only_b),
		len(identical) - len(different_name), len(different_name),
		len(similar)))

	# Report what has changed
	if only_a:
		print()
		print('## Only in A')
		for record in only_a:
			print('-', record.name)
	if only_b:
		print()
		print('## Only in B')
		for record in only_b:
			print('+', record.name)
	if different_name:
		print()
		print('## Different name (sequence identical)')
		for a, b in different_name:
			print('=', a.name, '--', b.name)
	if similar:
		print()
		print('## Similar')
		for a, b in similar:
			print_similar(a, b, colored)

	if has_duplicate_names or has_duplicate_sequences:
		sys.exit(2)
	if only_a or only_b or similar:
		sys.exit(1)
	# different name is fine for success
	sys.exit(0)



"""
Plot histograms of differences to reference V gene

For each gene, a histogram is plotted that shows how often a sequence was
assigned to that gene at a certain percentage difference.
"""
import sys
import logging
import numpy as np
import seaborn as sns
from matplotlib.backends.backend_pdf import FigureCanvasPdf, PdfPages
from matplotlib.figure import Figure

from .table import read_table

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--minimum-group-size', '-m', metavar='N', default=None, type=int,
		help="Plot only genes with at least N assigned sequences. "
		"Default: 0.1%% of assigned sequences or 100, whichever is smaller.")
	arg('--max-j-shm', metavar='VALUE', type=float, default=None,
		help='Use only rows with J%%SHM >= VALUE')
	arg('--multi', metavar='PDF', default=None,
		help='Plot individual error frequency histograms (for each V gene) to this PDF file')
	arg('--boxplot', metavar='PDF', default=None,
		help='Plot a single page with box(en)plots of V SHM for multiple genes')
	arg('table', metavar='FILTERED.TAB.GZ', help='Table with parsed IgBLAST results')


def plot_difference_histogram(group, gene_name, bins=np.arange(20.1)):
	"""
	Plot a histogram of percentage differences for a specific gene.
	"""
	exact_matches = group[group.V_SHM == 0]
	cdr3s_exact = len(set(s for s in exact_matches.CDR3_nt if s))
	js_exact = len(set(exact_matches.J_gene))

	fig = Figure(figsize=(100/25.4, 60/25.4))
	ax = fig.gca()
	ax.set_xlabel('Percentage difference')
	ax.set_ylabel('Frequency')
	fig.suptitle('Gene ' + gene_name, y=1.08, fontsize=16)
	ax.set_title('{:,} sequences assigned'.format(len(group)))

	ax.text(0.25, 0.95,
		'{:,} ({:.1%}) exact matches\n  {} unique CDR3\n  {} unique J'.format(
			len(exact_matches), len(exact_matches) / len(group),
			cdr3s_exact, js_exact),
		transform=ax.transAxes, fontsize=10,
		bbox=dict(boxstyle='round', facecolor='white', alpha=0.5),
		horizontalalignment='left', verticalalignment='top')
	_ = ax.hist(list(group.V_SHM), bins=bins)
	return fig


def main(args):
	table = read_table(args.table, usecols=['V_gene', 'J_gene', 'V_SHM', 'J_SHM', 'CDR3_nt'])
	if not args.multi and not args.boxplot:
		print('Don’t know what to do', file=sys.stderr)
		sys.exit(2)

	# Discard rows with any mutation within J at all
	logger.info('%s rows read', len(table))
	if args.max_j_shm is not None:
		# Discard rows with too many J mutations
		table = table[table.J_SHM <= args.max_j_shm][:]
		logger.info('%s rows remain after keeping only those with J%%SHM <= %s',
			len(table), args.max_j_shm)

	if args.minimum_group_size is None:
		total = len(table)
		minimum_group_size = min(total // 1000, 100)
		logger.info('Skipping genes with less than %s assignments', minimum_group_size)
	else:
		minimum_group_size = args.minimum_group_size

	# Genes with high enough assignment count
	all_genes = table['V_gene'].unique()
	genes = sorted(table['V_gene'].value_counts().loc[lambda x: x >= minimum_group_size].index)
	gene_set = set(genes)

	logger.info('%s out of %s genes have enough assignments', len(genes), len(all_genes))
	if args.multi:
		with PdfPages(args.multi) as pages:
			for gene, group in table.groupby('V_gene'):
				if gene not in gene_set:
					continue
				fig = plot_difference_histogram(group, gene)
				FigureCanvasPdf(fig).print_figure(pages, bbox_inches='tight')

		logger.info('Wrote %r', args.multi)

	if args.boxplot:
		aspect = 1 + len(genes) / 32
		g = sns.catplot(x='V_gene', y='V_SHM', kind='boxen', order=genes, data=table,
			height=2 * 2.54, aspect=aspect, color='g')
		# g.set(ylim=(-.1, None))
		g.set(ylabel='% V SHM (nt)')
		g.set(xlabel='V gene')
		g.set_xticklabels(rotation=90)
		g.savefig(args.boxplot)
		logger.info('Wrote %r', args.boxplot)



"""
Filter V gene candidates (germline and pre-germline filter)

After candidates for novel V genes have been found with the 'discover'
subcommand, this script is used to filter the candidates and make sure that
only true germline genes remain ("germline filter" and "pre-germline filter").
The following filtering and processing steps are performed on each candidate
separately:

* Discard sequences with N bases
* Discard sequences that come from a consensus over too few source sequences (unless whitelisted)
* Discard sequences with too few unique CDR3s (CDR3_clusters column)
* Discard sequences with too few unique Js (Js_exact column)
* Discard sequences identical to one of the database sequences (if DB given)
* Discard sequences that contain a stop codon (has_stop column) (unless whitelisted)

The following criteria involve comparison of candidates against each other:

* Discard sequences that are too similar to another (unless whitelisted)
* Discard sequences that are cross-mapping artifacts
* Discard sequences that have a too low allele ratio
* Discard sequences with too few unique Ds relative to other alleles (Ds_exact column)

If you provide a whitelist of sequences, then the candidates that appear on it
* are not checked for the cluster size criterion,
* are never considered near-duplicates,
* are allowed to contain a stop codon.

The filtered table is written to standard output.
"""
import sys
import logging
from collections import namedtuple
import pandas as pd
from sqt import FastaReader
from sqt.align import edit_distance

from .utils import UniqueNamer, is_same_gene, ChimeraFinder

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--cluster-size', type=int, metavar='N', default=0,
		help='Consensus must represent at least N sequences. '
		'Default: %(default)s')
	arg('--cross-mapping-ratio', type=float, metavar='RATIO', default=0.02,
		help='Ratio for detection of cross-mapping artifacts. Default: %(default)s')
	arg('--clonotype-ratio', '--allele-ratio', type=float, metavar='RATIO', default=0.1,
		help='Required ratio of "clonotypes" counts between alleles. '
			'Works only for genes named "NAME*ALLELE". Default: %(default)s')
	arg('--exact-ratio', type=float, metavar='RATIO', default=0.1,
		help='Required ratio of "exact" counts between alleles. '
			'Works only for genes named "NAME*ALLELE". Default: %(default)s')
	arg('--cdr3-shared-ratio', type=float, metavar='RATIO', default=1.0,
		help='Maximum allowed CDR3_shared_ratio. Default: %(default)s')
	arg('--minimum-db-diff', '-b', type=int, metavar='N', default=0,
		help='Sequences must have at least N differences to the database '
		'sequence. Default: %(default)s')
	arg('--maximum-N', '-N', type=int, metavar='COUNT', default=0,
		help='Sequences must have at most COUNT "N" bases. Default: %(default)s')
	arg('--unique-CDR3', '--CDR3s', type=int, metavar='N', default=1,
		help='Sequences must have at least N unique CDR3s within exact sequence matches. '
		'Default: %(default)s')
	# The default for unique-J is 0 because we might work on data without
	# any assigned J genes.
	arg('--unique-J', type=int, metavar='N', default=0,
		help='Sequences must have at least N unique Js within exact sequence matches. '
		'Default: %(default)s')
	arg('--unique-D-ratio', type=float, metavar='RATIO', default=None,
		help='Discard a sequence if another allele of this gene exists '
		'such that the ratio between their Ds_exact is less than RATIO')
	arg('--unique-D-threshold', type=int, metavar='THRESHOLD', default=10,
		help='Apply the --unique-D-ratio filter only if the Ds_exact of the other '
		'allele is at least THRESHOLD')
	arg('--allow-stop', action='store_true', default=False,
		help='Allow stop codons in sequences (uses the has_stop column).'
			'Default: Do not allow stop codons.')
	# arg('--allow-chimeras', action='store_true', default=False,
	# 	help='Do not filter out chimeric sequences. Default: Discard chimeras')
	arg('--whitelist', metavar='FASTA', default=[], action='append',
		help='Sequences that are never discarded or merged with others, '
			'even if criteria for discarding them would apply (except cross-mapping artifact '
			'removal, which is always performed).')
	arg('--fasta', metavar='FILE', help='Write new database in FASTA format to FILE')
	arg('--annotate', metavar='FILE',
		help='Write candidates.tab with filter annotations to FILE')
	arg('tables', metavar='CANDIDATES.TAB',
		help='Tables (one or more) created by the "discover" command',
		nargs='+')


Candidate = namedtuple('Candidate', ['sequence', 'name', 'clonotypes', 'exact', 'Ds_exact',
	'cluster_size', 'whitelisted', 'is_database', 'cluster_size_is_accurate', 'CDR3_start', 'row'])


class CrossMappingFilter:
	"""
	Filters a sequence if it is a cross-mapping artifact
	"""
	def __init__(self, cross_mapping_ratio: float):
		"""Check for cross-mapping"""
		self._ratio = cross_mapping_ratio

	def should_discard(self, a: Candidate, b: Candidate, dist: int, *args):
		"""
		Compare two candidates and decide which one should be filtered, if any.

		:param a: First candidate
		:param b: Second candidate
		:param dist: Edit distance between candidates
		:return: None if both candidates should be kept, a pair (x, reason) otherwise, where x is
		the candidate to be discarded (x is either a or b) and reason is a string describing why.
		"""
		if dist > 1:
			# Cross-mapping is unlikely if the edit distance is larger than 1
			return None
		if not a.is_database or not b.is_database:
			# Cross-mapping can only occur if both sequences are in the database
			return None

		total_count = (a.cluster_size + b.cluster_size)
		for u, v in [(a, b), (b, a)]:
			ratio = u.cluster_size / total_count
			if u.cluster_size_is_accurate and ratio < self._ratio:
				# u is probably a cross-mapping artifact of the higher-expressed v
				return u, f'xmap_of={v.name},xmap_ratio={ratio:.4f}'
		return None


class TooSimilarSequenceFilter:
	"""
	Filter out sequences that are too similar to another one
	"""
	@staticmethod
	def should_discard(a: Candidate, b: Candidate, dist: int, *args):
		if dist > 0:
			# The sequences are not similar, so keep both
			return None

		if a.whitelisted and b.whitelisted:
			# Both are whitelisted, so keep them
			return None

		if a.whitelisted:
			return b, f'identical_to={a.name},other_whitelisted'
		if b.whitelisted:
			return a, f'identical_to={b.name},other_whitelisted'

		# No sequence is whitelisted
		if b.clonotypes < a.clonotypes:
			return b, f'identical_to={a.name},fewer_clonotypes'
		# FIXME this is missing:
		# if a.clonotypes < b.clonotypes: ...
		if len(a.sequence) < len(b.sequence):
			return a, f'identical_to={b.name},shorter'
		return b, f'identical_to={a.name}'


class SameGeneFilter:
	def should_discard(self, a: Candidate, b: Candidate, dist: int, same_gene: bool):
		"""
		Compare two candidates and decide which one should be filtered, if any.

		:param a: First candidate
		:param b: Second candidate
		:param dist: Edit distance between candidates
		:return: None if both candidates should be kept, a pair (x, reason) otherwise, where x is
		the candidate to be discarded (x is either a or b) and reason is a string describing why.
		"""
		if not same_gene:
			# This filter applies only when comparing alleles of the same gene
			return None
		for u, v in [(a, b), (b, a)]:
			result = self.decide(a, b)
			if result is not None:
				return result
		return None

	def decide(self, a: Candidate, b: Candidate):
		raise NotImplementedError


class ClonotypeAlleleRatioFilter(SameGeneFilter):
	"""
	Clonotype allele ratio filter. Somewhat similar to cross-mapping, but
	this uses sequence names to decide whether two genes can be
	alleles of each other and the ratio is between the CDR3_clusters values
	"""
	def __init__(self, clonotype_ratio: float):
		self._ratio = clonotype_ratio

	def decide(self, u: Candidate, v: Candidate):
		if v.clonotypes == 0:
			return None
		ratio = u.clonotypes / v.clonotypes
		if ratio < self._ratio:
			# Clonotype allele ratio too low
			return u, f'clonotype_ratio={ratio:.4f},other={v.name}'
		return None


class ExactRatioFilter(SameGeneFilter):
	"""Exact V sequence occurrence allele ratio"""

	def __init__(self, exact_ratio: float):
		self._ratio = exact_ratio

	def decide(self, u: Candidate, v: Candidate):
		if v.exact == 0:
			return None
		ratio = u.exact / v.exact
		if ratio < self._ratio:
			# Allele ratio of exact occurrences too low
			return u, f'ex_occ_ratio={ratio:.1f},other={v.name}'
		return None


class UniqueDRatioFilter(SameGeneFilter):
	def __init__(self, unique_d_ratio: float, unique_d_threshold: int):
		self._unique_d_ratio = unique_d_ratio
		self._unique_d_threshold = unique_d_threshold

	def decide(self, u: Candidate, v: Candidate):
		if v.cluster_size < u.cluster_size or v.Ds_exact < self._unique_d_threshold:
			# TODO comment
			return None
		ratio = u.Ds_exact / v.Ds_exact
		if ratio < self._unique_d_ratio:
			# Ds_exact ratio too low
			return u, f'Ds_exact_ratio={ratio:.1f},other={v.name}'
		return None


class CandidateFilterer:
	"""
	Merge sequences that are sufficiently similar into single entries.
	"""
	def __init__(self, filters):
		self._items = []
		self._filters = filters

	def add(self, item):
		# This method could possibly be made simpler if the graph structure
		# was made explicit.
		items = []
		for existing_item in self._items:
			m = self.merged(existing_item, item)
			if m is None:
				items.append(existing_item)
			else:
				item = m
		items.append(item)
		self._items = items

	def extend(self, iterable):
		for i in iterable:
			self.add(i)

	def __iter__(self):
		if self._items and hasattr(self._items, 'name'):
			yield from sorted(self._items, key=lambda x: x.name)
		else:
			yield from self._items

	def __len__(self):
		return len(self._items)

	def merged(self, s: Candidate, t: Candidate):
		"""
		Given two candidates, decide whether to discard one of them and which one.

		Return None if both candidates should be kept.
		Return the candidate to keep otherwise.
		"""
		if len(s.sequence) > len(t.sequence):
			s, t = t, s  # make s always the shorter sequence

		# When computing edit distance between the two sequences, ignore the
		# bases in the 3' end that correspond to the CDR3
		s_no_cdr3 = s.sequence[:s.CDR3_start]
		t_no_cdr3 = t.sequence[:t.CDR3_start]
		if len(s_no_cdr3) != len(t_no_cdr3):
			t_prefix = t_no_cdr3[:len(s_no_cdr3)]
			t_suffix = t_no_cdr3[-len(s_no_cdr3):]
			dist_prefix = edit_distance(s_no_cdr3, t_prefix, 1)
			dist_suffix = edit_distance(s_no_cdr3, t_suffix, 1)
			dist_no_cdr3 = min(dist_prefix, dist_suffix)
		else:
			dist_no_cdr3 = edit_distance(s_no_cdr3, t_no_cdr3, 1)

		same_gene = is_same_gene(s.name, t.name)
		for filter_ in self._filters:
			result = filter_.should_discard(s, t, dist_no_cdr3, same_gene)
			if result is None:
				# not filtered
				continue
			if result[0] is s:
				return t
			else:
				return s

		# None of the filters decided to discard one of the sequences,
		# so keep both
		return None


class Whitelist:
	def __init__(self):
		self._sequences = dict()

	def add_fasta(self, path):
		with FastaReader(path) as fr:
			for record in fr:
				self._sequences[record.sequence.upper()] = record.name

	def closest(self, sequence):
		"""
		Search for the whitelist sequence that is closest to the given sequence.

		Return tuple (distance, name).
		"""
		if sequence in self._sequences:
			return 0, self._sequences[sequence]
		mindist = len(sequence)
		distances = []
		for seq, name in self._sequences.items():
			ed = edit_distance(seq, sequence, maxdiff=mindist)
			distances.append((ed, name))
			if ed == 1:
				# We know ed does not get smaller because the
				# 'sequence in whitelist' check
				# above covers that
				return ed, name
			mindist = min(mindist, ed)
		distance, name = min(distances)
		return distance, name

	def __len__(self):
		return len(self._sequences)

	def __contains__(self, other):
		return other in self._sequences


def is_chimera(table, whitelist):
	result = pd.Series('', index=table.index, dtype=object)
	if whitelist:
		whitelisted = table[table['whitelist_diff'] == 0]
	else:
		whitelisted = table[table['database_diff'] == 0]
	chimera_finder = ChimeraFinder(list(whitelisted['consensus']))
	for row in table[(table['whitelist_diff'] != 0) & (table['database_diff'] != 0)].itertuples():
		query = row.consensus
		chimera_result = chimera_finder.find_exact(query)
		if chimera_result:
			(prefix_length, prefix_indices, suffix_indices) = chimera_result
			suffix_length = len(query) - prefix_length
			prefix_name = whitelisted.iloc[prefix_indices[0]]['name']
			suffix_row = whitelisted.iloc[suffix_indices[0]]
			suffix_name = suffix_row['name']
			suffix_sequence = suffix_row['consensus']
			# logger.info('Candidate %s (diffs.: %d) appears to be a chimera of '
			# 	'%s:1..%d and %s:%d..%d', row.name, row.whitelist_diff, prefix_name, prefix_length,
			# 	suffix_name, len(suffix_sequence) - suffix_length + 1, len(suffix_sequence))
			result.loc[row.Index] = '{}:1..{}+{}:{}..{}'.format(prefix_name, prefix_length,
				suffix_name, len(suffix_sequence) - suffix_length + 1, len(suffix_sequence))

	return result


def mark_rows(table, condition, reason):
	table.loc[condition, 'why_filtered'] += reason + ';'
	table.loc[condition, 'is_filtered'] += 1
	logger.info('Marked %s candidates as %r', sum(condition), reason)


def main(args):
	if args.unique_D_threshold <= 1:
		sys.exit('--unique-D-threshold must be at least 1')
	filters = []
	if args.cross_mapping_ratio:
		filters.append(CrossMappingFilter(args.cross_mapping_ratio))
	if args.clonotype_ratio:
		filters.append(ClonotypeAlleleRatioFilter(args.clonotype_ratio))
	if args.exact_ratio:
		filters.append(ExactRatioFilter(args.exact_ratio))
	if args.unique_D_ratio or args.unique_D_threshold:
		filters.append(UniqueDRatioFilter(args.unique_D_ratio, args.unique_D_threshold))
	filters.append(TooSimilarSequenceFilter())
	merger = CandidateFilterer(filters)
	whitelist = Whitelist()
	for path in args.whitelist:
		whitelist.add_fasta(path)
	logger.info('%d unique sequences in whitelist', len(whitelist))

	# Read in tables
	total = 0
	overall_table = None
	for path in args.tables:
		table = pd.read_csv(path, sep='\t')
		i = list(table.columns).index('consensus')
		# whitelist_diff distinguishes between 0 and !=0 only
		# at this point. Accurate edit distances are computed later.
		whitelist_diff = [(0 if s in whitelist else -1) for s in table['consensus']]
		# TODO rename to is_whitelisted
		table.insert(i, 'whitelist_diff', pd.Series(whitelist_diff, index=table.index, dtype=int))
		table.insert(i+1, 'closest_whitelist', pd.Series('', index=table.index))
		table.insert(3, 'why_filtered', pd.Series('', index=table.index))
		table.insert(3, 'is_filtered', pd.Series(0, index=table.index))

		mark_rows(table, table.database_diff < args.minimum_db_diff, 'too_low_dbdiff')

		if 'N_bases' in table.columns:
			mark_rows(table, table.N_bases > args.maximum_N, 'too_many_N_bases')
		mark_rows(table, table.CDR3s_exact < args.unique_CDR3, 'too_low_CDR3s_exact')
		mark_rows(table, table.CDR3_shared_ratio > args.cdr3_shared_ratio, 'too_high_CDR3_shared_ratio')
		mark_rows(table, table.Js_exact < args.unique_J, 'too_low_Js_exact')
		if not args.allow_stop:
			mark_rows(table, (table.has_stop != 0) & (table.whitelist_diff != 0), 'has_stop')
		mark_rows(table, (table.cluster_size < args.cluster_size) & (table.whitelist_diff != 0),
			'too_low_cluster_size')
		table['database_changes'].fillna('', inplace=True)
		logger.info('Table read from %r contains %s candidate V gene sequences. '
			'%s remain after per-entry filtering', path,
			len(table), sum(table.is_filtered == 0))
		if args.whitelist:
			logger.info('Of those, %d are protected by the whitelist', sum(table.whitelist_diff == 0))
		total += len(table)

		if overall_table is None:
			overall_table = table
		else:
			overall_table.append(table)
	del table
	if len(args.tables) > 1:
		logger.info('Read %s tables with %s entries total. '
			'After per-entry filtering, %s entries remain.', len(args.tables),
			len(overall_table), sum(overall_table.is_filtered == 0))

	def cluster_size_is_accurate(row):
		return bool(set(row.cluster.split(';')) & {'all', 'db'})

	for _, row in overall_table.iterrows():
		if row['is_filtered'] > 0:
			continue
		merger.add(Candidate(
			sequence=row['consensus'],
			name=row['name'],
			clonotypes=row['clonotypes'],
			exact=row['exact'],
			Ds_exact=row['Ds_exact'],
			cluster_size=row['cluster_size'],
			whitelisted=row['whitelist_diff'] == 0,
			is_database=row['database_diff'] == 0,
			cluster_size_is_accurate=cluster_size_is_accurate(row),
			CDR3_start=row.get('CDR3_start', 10000),  # TODO backwards compatibility
			row=row.name,  # row.name is the index of the row. It is not row['name'].
		))

	# Discard near-duplicates
	overall_table.loc[overall_table.is_filtered > 0, 'is_duplicate'] = False
	overall_table.loc[overall_table.is_filtered == 0, 'is_duplicate'] = True
	for info in merger:
		overall_table.loc[info.row, 'is_duplicate'] = False
	mark_rows(overall_table, overall_table.is_duplicate, 'is_duplicate')
	del overall_table['is_duplicate']

	# Name sequences
	overall_table['name'] = overall_table['name'].apply(UniqueNamer())
	overall_table.sort_values(['name'], inplace=True)

	# Because whitelist_dist() is expensive, this is run when
	# all of the filtering has already been done
	if whitelist:
		for row in overall_table.itertuples():
			# TODO skipping this is just a performance optimization
			if row.is_filtered > 0:
				continue
			distance, name = whitelist.closest(overall_table.loc[row[0], 'consensus'])
			overall_table.loc[row[0], 'closest_whitelist'] = name
			overall_table.loc[row[0], 'whitelist_diff'] = distance
	else:
		overall_table.whitelist_diff.replace(-1, '', inplace=True)

	i = list(overall_table.columns).index('database_diff')
	# TODO
	# chimeras are computed for the full table, not the filtered one. what should be done?
	overall_table.insert(i, 'chimera', is_chimera(overall_table, whitelist))

	# Discard chimeric sequences
	# if not args.allow_chimeras:
	#   overall_table = overall_table[~is_chimera(overall_table, whitelist)].copy()

	filtered_table = overall_table[overall_table.is_filtered == 0]
	del filtered_table['is_filtered']
	del filtered_table['why_filtered']
	print(filtered_table.to_csv(sep='\t', index=False, float_format='%.2f'), end='')

	if args.annotate:
		with open(args.annotate, 'w') as f:
			print(overall_table.to_csv(sep='\t', index=False, float_format='%.2f'), end='', file=f)
	if args.fasta:
		with open(args.fasta, 'w') as f:
			for _, row in filtered_table.iterrows():
				print('>{}\n{}'.format(row['name'], row['consensus']), file=f)

	logger.info('%d sequences in new database', len(filtered_table))



#!/usr/bin/env python3
"""
Extract V_SHM values from a filtered.tab.gz and group them by V gene

Create a table that has a column for each V gene.
"""
import logging
import pandas as pd


def add_arguments(parser):
	arg = parser.add_argument
	arg('--field', default='V_SHM', help='Which column to extract. Default: %(default)s')
	arg('table', help='filtered.tab.gz file')


def main(args):
	table = pd.read_table(args.table, sep='\t', usecols=(args.field, 'V_gene'))
	df = pd.DataFrame(index=range(table.groupby('V_gene').size().max()))
	for gene, group in table.groupby('V_gene'):
		df[gene] = group[args.field].reset_index(drop=True)
	print(df.to_csv(sep='\t', index=False))


if __name__ == '__main__':
	from argparse import ArgumentParser
	parser = ArgumentParser()
	add_arguments(parser)
	args = parser.parse_args()
	main(args)



"""
Draw a dendrogram of sequences in a FASTA file.
"""
import logging
import numpy as np
import matplotlib
from scipy.spatial import distance
from scipy.cluster import hierarchy
from sqt import FastaReader
from igdiscover.utils import distances


logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	arg('--mark', '--db', metavar='FASTA',
		help='Path to a FASTA file with a set of "known" sequences. Sequences '
		'in the main file that do *not* occur here will be marked with (new). '
		'If not given, no sequences will be marked (use this to compare two '
		'databases).')
	arg('--method', choices=('single', 'complete', 'weighted', 'average'),
		default='average',
		help='Linkage method. Default: "average" (=UPGMA)')
	arg('fasta', help='Path to input FASTA file')
	arg('plot', help='Path to output PDF or PNG')


class PrefixComparer:
	def __init__(self, sequences):
		self._sequences = [ s.upper() for s in sequences ]

	def __contains__(self, other):
		for seq in self._sequences:
			if seq.startswith(other) or other.startswith(seq):
				return True
		return False


def main(args):
	with FastaReader(args.fasta) as fr:
		sequences = list(fr)
	logger.info('Plotting dendrogram of %s sequences', len(sequences))
	if args.mark:
		with FastaReader(args.mark) as fr:
			mark = PrefixComparer(record.sequence for record in fr)
		labels = []
		n_new = 0
		for record in sequences:
			if record.sequence not in mark:
				extra = ' (new)'
				n_new += 1
			else:
				extra = ''
			labels.append(record.name + extra)
		logger.info('%s sequence(s) marked as "new"', n_new)
	else:
		labels = [s.name for s in sequences]

	import seaborn as sns
	import matplotlib.pyplot as plt
	sns.set()
	sns.set_style("white")
	font_size = 297 / 25.4 * 72 / (len(labels) + 5)
	font_size = min(16, max(6, font_size))
	height = font_size * (len(labels) + 5) / 72
	fig = plt.figure(figsize=(210 / 25.4, height))
	matplotlib.rcParams.update({'font.size': 4})
	ax = fig.gca()
	sns.despine(ax=ax, top=True, right=True, left=True, bottom=True)
	sns.set_style('whitegrid')
	if len(sequences) >= 2:
		m = distances([s.sequence for s in sequences])
		y = distance.squareform(m)
		mindist = int(y.min())
		logger.info('Smallest distance is %s. Found between:', mindist)
		for i,j in np.argwhere(m == y.min()):
			if i < j:
				logger.info('%s and %s', labels[i], labels[j])
		l = hierarchy.linkage(y, method=args.method)
		hierarchy.dendrogram(l, labels=labels, leaf_font_size=font_size, orientation='right', color_threshold=0.95*max(l[:,2]))
	else:
		ax.text(0.5, 0.5, 'no sequences', fontsize='xx-large')
	ax.grid(False)
	fig.set_tight_layout(True)
	fig.savefig(args.plot)



#!/usr/bin/env python3
"""
Merge overlapping sequences
"""
import sys
from argparse import ArgumentParser
from sqt.io.fasta import FastaReader
from cutadapt.align import Aligner


class Merger:
	"""
	Merge similar items into one. To specify what "similar" means, implement
	the merged() method in a subclass.
	"""
	def __init__(self):
		self._items = []

	def add(self, item):
		# This method could possibly be made simpler if the graph structure
		# was made explicit.

		items = []
		for existing_item in self._items:
			m = self.merged(existing_item, item)
			if m is None:
				items.append(existing_item)
			else:
				item = m
		items.append(item)
		self._items = items

	def extend(self, iterable):
		for i in iterable:
			self.add(i)

	def __iter__(self):
		if self._items and hasattr(self._items, 'name'):
			yield from sorted(self._items, key=lambda x: x.name)
		else:
			yield from self._items

	def __len__(self):
		return len(self._items)

	def merged(self, existing_item, item):
		"""
		If existing_item and item can be returned, this method must return
		a new item that represents a merged version of both. If they cannot
		be merged, it must return None.
		"""
		raise NotImplementedError("not implemented")


def find_overlap(s, t, min_overlap=1):
	"""
	Detect if s and t overlap.

	Returns:

	None if no overlap was detected.
	0 if s is a prefix of t or t is a prefix of s.
	Positive int gives index where t starts within s.
	Negative int gives -index where s starts within t.

	>>> find_overlap('ABCDE', 'CDE')
	2
	>>> find_overlap('CDE', 'ABCDEFG')
	-2
	>>> find_overlap('ABC', 'X') is None
	True
	"""
	aligner = Aligner(s, max_error_rate=0)
	aligner.min_overlap = min_overlap
	result = aligner.locate(t)
	if result is None:
		return None
	s_start, _, t_start, _, _, _ = result
	return s_start - t_start


def merge_overlapping(s, t):
	"""
	Return merged sequences or None if they do not overlap
	The minimum overlap is 50% of the length of the shorter sequence.
	"""
	i = find_overlap(s, t, min_overlap=max(1, min(len(s), len(t)) // 2))
	if i is None:
		return None

	if i >= 0:
		# positive: index of t in s
		if i + len(t) < len(s):
			# t is in s
			return s
		return s[:i] + t
	if -i + len(s) < len(t):
		# s is in t
		return t
	return t[:-i] + s


class SequenceInfo:
	__slots__ = ('name', 'sequence', 'count')

	def __init__(self, name, sequence):
		self.name = name
		self.sequence = sequence
		# self.count = count


class OverlappingSequenceMerger(Merger):
	"""
	Merge sequences that overlap
	"""
	def merged(self, s, t):
		"""
		Merge two sequences if they overlap. Return None if they should not be merged.
		"""
		m = merge_overlapping(s.sequence, t.sequence)
		if m is not None:
			# overlap found
			return SequenceInfo(s.name + ';' + t.name, m)

		# no overlap found
		return None


def add_arguments(parser):
	arg = parser.add_argument
	arg('fasta', help='File with sequences')


def main(args):
	merger = OverlappingSequenceMerger()
	with FastaReader(args.fasta) as f:
		for record in f:
			merger.add(SequenceInfo(record.name, record.sequence))

	for info in merger:
		print('>{}\n{}'.format(info.name, info.sequence))


if __name__ == '__main__':
	parser = ArgumentParser(description=__doc__)
	add_arguments(parser)
	args = parser.parse_args()
	main(args)



"""
Query two ...
"""
import logging
from collections import defaultdict
from contextlib import ExitStack

import pandas as pd
from xopen import xopen

from igdiscover.table import read_table
from igdiscover.utils import slice_arg
from igdiscover.clonotypes import is_similar_with_junction, CLONOTYPE_COLUMNS

logger = logging.getLogger(__name__)


def add_arguments(parser):
	arg = parser.add_argument
	# arg('--minimum-count', '-c', metavar='N', default=1, type=int,
	# 	help='Discard all rows with count less than N. Default: %(default)s')
	# arg('--cdr3-core', default=None,
	# 	type=slice_arg, metavar='START:END',
	# 	help='START:END defines the non-junction region of CDR3 '
	# 		'sequences. Use negative numbers for END to count '
	# 		'from the end. Regions before and after are considered to '
	# 		'be junction sequence, and for two CDR3s to be considered '
	# 		'similar, at least one of the junctions must be identical. '
	# 		'Default: no junction region.')
	# arg('--mismatches', default=1, type=float,
	# 	help='No. of allowed mismatches between CDR3 sequences. '
	# 		'Can also be a fraction between 0 and 1 (such as 0.15), '
	# 		'interpreted relative to the length of the CDR3 (minus the front non-core). '
	# 		'Default: %(default)s')
	arg('--aa', default=False, action='store_true',
		help='Count CDR3 mismatches on amino-acid level. Default: Compare nucleotides.')
	arg('tables', nargs='+', help='clonotype tables')


def collect(querytable, reftable, mismatches, cdr3_core_slice, cdr3_column):
	"""
	Find all queries from the querytable in the reftable.

	Yield tuples (query_rows, similar_rows) where the query_rows is a list
	with all the rows that have the same result. similar_rows is a DataFrame
	whose rows are the ones matching the query.
	"""
	# The vjlentype is a "clonotype without CDR3 sequence" (only V, J, CDR3 length)
	# Determine set of vjlentypes to query
	query_vjlentypes = defaultdict(list)
	for row in querytable.itertuples():
		vjlentype = (row.V_gene, row.J_gene, len(row.CDR3_nt))
		query_vjlentypes[vjlentype].append(row)

	groupby = ('V_gene', 'J_gene', 'CDR3_length')
	for vjlentype, vjlen_group in reftable.groupby(groupby):
		# (v_gene, j_gene, cdr3_length) = vjlentype
		if vjlentype not in query_vjlentypes:
			continue

		# Collect results for this vjlentype. The results dict
		# maps row indices (into the vjlen_group) to each query_row,
		# allowing us to group identical results together.
		results = defaultdict(list)
		for query_row in query_vjlentypes.pop(vjlentype):
			cdr3 = getattr(query_row, cdr3_column)
			# Save indices of the rows that are similar to this query
			indices = tuple(index for index, r in enumerate(vjlen_group.itertuples())
				if is_similar_with_junction(cdr3, getattr(r, cdr3_column), mismatches, cdr3_core_slice))
			results[indices].append(query_row)

		# Yield results, grouping queries that lead to the same result
		for indices, query_rows in results.items():
			if not indices:
				for query_row in query_rows:
					yield ([query_row], [])
				continue

			similar_group = vjlen_group.iloc[list(indices), :]
			yield (query_rows, similar_group)

	# Yield result tuples for all the queries that have not been found
	for queries in query_vjlentypes.values():
		for query_row in queries:
			yield ([query_row], [])


def main(args):
	cdr3_column = 'CDR3_aa' if args.aa else 'CDR3_nt'
	# CDR3_length is unused, but this ensures that we are reading the right type of file
	usecols = ['count', 'V_gene', 'J_gene', 'CDR3_nt', 'CDR3_length', cdr3_column]
	tables = []
	for path in args.tables:
		table = read_table(path, usecols=usecols).set_index(
			['V_gene', 'J_gene', cdr3_column])
		tables.append(table)
		logger.info('Read table from %r with %s rows', path, len(table))

	common = tables[0].join(tables[1], lsuffix='_left', rsuffix='_right', how='inner')
	print('Common clonotypes:', len(common))

	n = common[['count_left', 'count_right']].min(axis=1).sum()
	print('Common sequences:', n)


if __name__ == '__main__':
	from argparse import ArgumentParser
	parser = ArgumentParser()
	add_arguments(parser)
	args = parser.parse_args()
	main(args)
